/*
** pci_scan
**
** Copyright James A. McQuillan (jam@McQuil.com), 2001
**
** This program is under the GNU General Public License (GPL)
** For the complete license, see the included 'COPYING' file.
**
** Purpose:
**
**    This program will read in a device database of PCI Vendor/Device Ids.
**    It will then check the /proc/bus/pci/devices file, to see if any
**    devices in the system are found in the list that was read in.
**
**    The device database file will contain records that look
**    like this:
**
**        1234:5678  somedriver
**
**    The device database file will be read into a linked list
**    in memory.
**    The program will then read each entry in the /proc/bus/pci/devices
**    file.  For each entry, it will search the linked list to see if
**    it finds a match.  As soon as a match is found, the drivername
**    will be printed to stdout, and the program will exit.
**
**    This is very useful, to determine what kind of network card
**    or video card is installed in the system, and which driver
**    module should be loaded.
**
** Usage:
**
**    pci_scan /etc/niclist
**
** Example:
**
**    NIC=`pci_scan /etc/niclist`
**
*/

#include <stdio.h>
#include <errno.h>
#include <malloc.h>

#define TRUE    1
#define FALSE   0


// Function prototypes

void trim(char *);

//
// Globals
//

#define RECSIZE 10240
#define FALSE   0
#define TRUE    1

char sInBuf[RECSIZE];

#define PROCFILE "/proc/bus/pci/devices"

char pci_database[64];

struct pci_dev {
    char pci_id[10];
    char driver[64];
    struct pci_dev  *next;
};

typedef struct pci_dev PCI_DEV;

PCI_DEV  *first = (PCI_DEV *)NULL;
PCI_DEV  *next  = (PCI_DEV *)NULL;
PCI_DEV  *prev  = (PCI_DEV *)NULL;

/*------------------------------------------------------------------------------
**
** void trim(char *)
**
** This routine will trim leading and trailing white-space from
** a character string.  It will also strip out comments that start
** with a hash '#' mark.
**
** It modifies the string that is passed to it.
**
*/

void trim(char *str)
{
    char *p;
    char *tmpbuf;
    char *q;
    int  nQuoteCount=0;

    // Trim leading spaces

    while( isspace(*str) ){
        p = str;
        while(*p){
            *p = *(p+1);
            p++;
        }
    }

    // Look for hash '#' characters to denote comments
    // If we find one, and it is NOT inside of a quoted string
    // then replace it with a NULL to terminate the string.

    p = str;
    while(*p){
        if(*p == '\"')
            nQuoteCount++;
        if( *p == '#' && ( nQuoteCount % 2 ) == 0 )
            *p = '\0';      // replace '#' with NULL
        else
            p++;
    }

    // Coming out of the previous loop, p points to the terminating NULL


    // Trim trailing spaces

    p--;                    // Get off of the trailing NULL

    while( p >= str && isspace(*p) ){
        *p = '\0';          // Replace the space with a NULL
        p--;
    }

    // Ok, now we need to remove quotes.  We allocate a chunk of
    // memory and copy the string into the new space.
    tmpbuf = (char *)malloc(strlen(str)+1);
    strcpy(tmpbuf,str);

    p = tmpbuf;
    q = str;
    while(*p){
        if( *p != '\"' )
            *q++ = *p;
        p++;
    }
    *q = '\0';

    free(tmpbuf);
    
}

/*------------------------------------------------------------------------------
*/
load_pci_database(){

    FILE *pcifile;
    int  fEof = FALSE;

    if( ( pcifile = fopen( pci_database, "r" ) ) == NULL ){
        char buf[64];
        sprintf( buf, "Open of %s failed", pci_database );
        perror(buf);
        exit(1);
    }

    while( !fEof ){
        if( fgets( sInBuf, RECSIZE-1, pcifile ) == NULL ){
            fEof = TRUE;
        }
        else{
            char sPciInfo[64];
            char sDriverName[64];
            int  nCnt;
            trim(sInBuf);
            if( strlen( sInBuf ) > 0 ){
                nCnt = sscanf( sInBuf, "%s %s", sPciInfo, sDriverName );
                if( nCnt == 2 ){
                    prev = next;
                    
                    next = (PCI_DEV *)malloc(sizeof(PCI_DEV));
                    if( first == (PCI_DEV *)NULL )
                        first = next;
                    next->next = (PCI_DEV *)NULL;

                    strcpy( next->pci_id, sPciInfo );
                    strcpy( next->driver, sDriverName );
                    if( prev )
                        prev->next = next;
                }
            }
        }
    }

    fclose(pcifile);
}

/*------------------------------------------------------------------------------
*/

main( int argc, const char **argv ){

    FILE *procfile;

    int  fEof      = FALSE;
    int  fFoundDev = FALSE;

    if( argc < 2 ){
        fprintf( stderr,"Usage:  %s pcilist\n", argv[0] );
        exit(1);
    }

    strcpy( pci_database, argv[1] );

    load_pci_database();

    if( ( procfile = fopen( PROCFILE, "r" ) ) == NULL ){
        char buf[64];
        sprintf(buf,"Open of %s failed", PROCFILE);
        perror(buf);
	exit(1);
    }

    while( !fEof && !fFoundDev ){
        if( fgets(sInBuf, RECSIZE-1, procfile ) == NULL ){
            fEof = TRUE;
        }
        else{
            char sBusInfo[64];
            char sPciInfo[64];
            int  nCnt;
            nCnt = sscanf( sInBuf, "%s %s", sBusInfo, sPciInfo );
            if(nCnt == 2){
                int idx = 0;
                int i;
                //
                // change an entry that looks like this:
                //      12345678
                // We need to change it to this:
                //      1234:5678
                //
                sPciInfo[9] = '\0';
                for( i = 7; i >= 4; i-- ){
                    sPciInfo[i+1] = sPciInfo[i];
                }
                sPciInfo[4] = ':';
		next = first;
		// fprintf(stderr,"sPciInfo = %s\n", sPciInfo );
		while( next && !fFoundDev ){
                    if( strcasecmp(next->pci_id, sPciInfo ) == 0 ){
                        printf("%s\n",next->driver);
                        fFoundDev = TRUE;
                    }
                    next = next->next;
                }
            }
        }
    }

    fclose(procfile);

    next = first;
    while( next ){
        PCI_DEV *p;
        p = next;
        next = next->next;
        free((PCI_DEV *)p);
    }
    return(0);
}

/*------------------------------------------------------------------------------
*/
